@@ -10,19 +10,26 @@ import os
 
 # isort: off
 
-from .version import __version__
+from .version import __version__ # noqa
 
 # isort: on
 
-from .distance import *  # noqa: F403
-from .model import MorpheusModel, MorpheusModels
-from .sumstat import *  # noqa: F403
+from . import C, log # noqa
+from .distance import KolmogorovSmirnovDistance # noqa
+from .model import MorpheusModel, MorpheusModels # noqa
+from .sumstat import ( # noqa
+    CCContributorsAllTpCountSumstatFun,
+    CellCountSumstatFun,
+    ClusterCountSumstatFun,
+    SummaryStatistics,
+)
 
 # Set log level
 try:
     loglevel = os.environ['FitMultiCell_LOG_LEVEL'].upper()
 except KeyError:
     loglevel = 'INFO'
+
 logger = logging.getLogger("FitMultiCell")
 logger.setLevel(loglevel)
 sh = logging.StreamHandler()
