@@ -1,26 +1,15 @@
 import copy
 import logging
-import numbers
 import os
 import shutil
 import xml.etree.ElementTree as ET  # noqa: S405
-from types import ModuleType
-from typing import Callable, List, Union
-
-import numpy as np
-import pandas as pd
-
-from ..sumstat import IdSumstatFun
-from ..util import base as util_base
-
+from typing import Callable, Dict, Sequence, Union
+from pyabc import Parameter
+from pyabc.external import LOC, TIMEOUT, ExternalModel
+from .. import util, C
+from ..sumstat import SummaryStatistics
 logger = logging.getLogger("FitMultiCell.Model")
 
-try:
-    from pyabc.external.base import ExternalModel
-    from pyabc.parameters import Parameter
-except ImportError:
-    logger.error("pyabc must be installed.")
-
 
 class MorpheusModel(ExternalModel):
     """
@@ -30,14 +19,13 @@ class MorpheusModel(ExternalModel):
 
     Parameters
     ----------
-
     model_file:
         The XML file containing the morpheus model.
     par_map:
-        A dictionary from str to str, the keys being the parameter ids
+        A dictionary from string to string, the keys being the parameter ids
         to be used in pyabc, and the values xpaths in the `morpheus_file`.
     par_scale:
-        A dictionary or str to state the scale used to define the parameter
+        A dictionary or string to state the scale used to define the parameter
         space, e.g., lin, log10, log2
     sumstat_funs:
         List of functions to calculate summary statistics. The list entries
@@ -51,6 +39,8 @@ class MorpheusModel(ExternalModel):
         Directory to put the temporary folders into. The default is
         the system's temporary files location. Note that these files
         are usually deleted upon system shutdown.
+    clean_simulation:
+        Whether to remove simulation files when they are no longer needed.
     show_stdout, show_stderr:
         Whether to show or hide the stdout and stderr streams.
     raise_on_error:
@@ -66,7 +56,7 @@ class MorpheusModel(ExternalModel):
         solve the issue with result that cannot be eliminated from morpheus
         output but yet are not used in the fitting process.
     timeout:
-        A time in seconds used to early stop Morpheus simulation when exceed.
+        Maximum execution time in seconds, after which Morpheus is stopped.
     ss_post_processing:
         A callable function to perform post processing on Morpheus output. If
         a dict is passed, then specific function will be applied to each
@@ -78,11 +68,12 @@ class MorpheusModel(ExternalModel):
     def __init__(
         self,
         model_file: str,
-        par_map: dict,
-        par_scale: Union[dict, str] = "lin",
-        exp_cond_map: dict = None,
-        sumstat_funs: List = None,
-        executable: str = "morpheus",
+        par_map: Dict[str, str],
+        sumstat: SummaryStatistics = None,
+        par_scale: Union[Dict[str, str], str] = C.LIN,
+        exp_cond_map: Dict = None,
+        executable: str = C.MORPHEUS,
+        gui_executable: str = C.GUI_MORPHEUS,
         suffix: str = None,
         prefix: str = "morpheus_model_",
         dir: str = None,
@@ -90,12 +81,12 @@ class MorpheusModel(ExternalModel):
         show_stdout: bool = False,
         show_stderr: bool = True,
         raise_on_error: bool = False,
+        timeout: float = None,
         name: str = None,
-        time_var: str = "time",
-        ignore_list: list = None,
-        timeout: float = np.inf,
+        time_var: str = C.TIME,
+        outputdir: str = None,
         ss_post_processing: Union[Callable, dict] = None,
-        output_file: str = "logger.csv",
+        # output_file: str = C.OUTPUT_FILE,
     ):
         if name is None:
             name = model_file
@@ -107,43 +98,50 @@ class MorpheusModel(ExternalModel):
             suffix=suffix,
             prefix=prefix,
             dir=dir,
-            clean_simulation=clean_simulation,
             show_stdout=show_stdout,
             show_stderr=show_stderr,
             raise_on_error=raise_on_error,
+            timeout=timeout,
             name=name,
         )
+        self.gui_executable = gui_executable
+        self.clean_simulation: bool = clean_simulation
 
-        self.par_map = par_map
-        self.par_scale = par_scale
-        self.exp_cond_map = exp_cond_map
-        if sumstat_funs is None:
-            if self.exp_cond_map is None:
-                sumstat_funs = [IdSumstatFun()]
-            else:
-                sumstat_funs = [
-                    IdSumstatFun(name=list(self.exp_cond_map.keys())[0])
-                ]
-        self.clean_simulation = clean_simulation
-        self.sumstat_funs = sumstat_funs
-        self._check_sumstat_funs()
-        self.time_var = time_var
-        if ignore_list is None:
-            ignore_list = []
-        self.ignore_list = ignore_list
-        self.timeout = timeout
-        self.ss_post_processing = ss_post_processing
-        self.output_file = output_file
-        self._sanity_chceck_model()
+        self.par_map: Dict[str, str] = par_map
+
+        if isinstance(par_scale, str):
+            par_scale = {key: par_scale for key in par_map}
+        self.par_scale: Dict[str, str] = par_scale
+
+        self.exp_cond_map: Dict = exp_cond_map
+        self.timeout: float = timeout
+        # if sumstat_funs is None:
+        #     if self.exp_cond_map is None:
+        #         sumstat_funs = [IdSumstatFun()]
+        #     else:
+        #         sumstat_funs = [
+        #             IdSumstatFun(name=list(self.exp_cond_map.keys())[0])
+        #         ]
+        # self.summary_statistics()
+        self.sumstat = sumstat
+        # if sumstat is not None:
+        #     self._check_sumstat_funs()
+        if self.exp_cond_map is not None:
+            self.sumstat.name = list(self.exp_cond_map.keys())[0]
+        self.clean_simulation: bool = clean_simulation
+        self.time_var: str = time_var
+        self.outputdir = outputdir
+        self.ss_post_processing: Union[Callable, dict] = ss_post_processing
+        # self.output_file: str = output_file
 
     def __str__(self):
         s = (
             f"MorpheusModel {{\n"
-            f"\tname      : {self.name}\n"
-            f"\tpar_map      : {self.par_map}\n"
-            f"\tsumstat_funs      : {self.sumstat_funs}\n"
-            f"\texecutable      : {self.eh.executable}\n"
-            f"\toutput_file      : {self.output_file}\n"
+            f"  name         : {self.name}\n"
+            f"  par_map      : {self.par_map}\n"
+            f"  sumstat      : {self.sumstat}\n"
+            f"  executable   : {self.eh.executable}\n"
+            # f"  output_file  : {self.output_file}\n"
             f"}}"
         )
         return s
@@ -152,36 +150,36 @@ class MorpheusModel(ExternalModel):
         return self.__str__()
 
     def __call__(self, pars: Parameter):
-        """
+        """Simulate data for parameters.
+
         This function is used in ABCSMC (or rather the sample() function,
         which redirects here) to simulate data for given parameters `pars`.
         """
-
-        # TODO: move all the content to another function and just call here.
-
         # create target on file system
+        if self.outputdir is not None:
+            self.eh.dir = self.outputdir
         loc = self.eh.create_loc()
-        file_ = os.path.join(loc, "model.xml")
+        model_file = os.path.join(loc, "model.xml")
 
         # write new file with parameter modifications
-        self.write_modified_model_file(file_, pars)
+        self.write_modified_model_file(model_file, pars)
 
         # create command
         cmd = self.eh.create_executable(loc)
-        self.eh.timeout = self.timeout
-        cmd = cmd + f" -file={file_} -outdir={loc}"
-
+        cmd = cmd + f" -file={model_file} -outdir={loc}"
+        if self.eh.timeout is None and self.timeout is not None:
+            self.eh.timeout = self.timeout
         # call the model
-        status = self.eh.run(args=pars, cmd=cmd, loc=loc)
+        status = self.eh.run(cmd=cmd, loc=loc)
+
+        # check whether simulation timed out
+        if status["returncode"] == TIMEOUT:
+            # remove simulation output
+            if self.clean_simulation:
+                clean_simulation_output(loc)
+            return TIMEOUT
 
         # compute summary statistics
-        if status == -15:
-            msg = (
-                f"Simulation exceed time limit limit: {self.timeout}s "
-                f"for arguments {pars}: returncode {status}."
-            )
-            logger.warning(msg)
-            return -15
         sumstats = self.compute_sumstats(loc)
 
         # remove simulation output
@@ -189,7 +187,8 @@ class MorpheusModel(ExternalModel):
             clean_simulation_output(loc)
 
         # perform data post-process on Morpheus output
-        sumstats = self.call_post_processing_ss(sumstats)
+        # sumstats = self.call_post_processing_ss(sumstats)
+
         return sumstats
 
     def get_parmap_xpath_attr(self, key, attrib='value'):
@@ -202,9 +201,7 @@ class MorpheusModel(ExternalModel):
             name of parameter of interest.
         attrib: str
             the type of attribute that need to be changed on the xml file.
-
         """
-
         par = self.par_map[key]
         if isinstance(par, str):
             return par, attrib
@@ -225,9 +222,7 @@ class MorpheusModel(ExternalModel):
             name of experimental condition of interest.
         attrib: str
             the type of attribute that need to be changed on the xml file.
-
         """
-
         exp_cond = self.exp_cond_map[key]
         if isinstance(exp_cond, str):
             return exp_cond, attrib
@@ -238,13 +233,12 @@ class MorpheusModel(ExternalModel):
                 f"par_map[{key}] should be a str or a list/tuple of length 2"
             )
 
-    def write_modified_model_file(self, file_, pars):
+    def write_modified_model_file(self, file_: str, pars: Dict[str, float]):
         """
         Write a modified version of the morpheus xml file to the target
         directory.
         """
-        temp_par = copy.deepcopy(pars)
-        rescaled_pars = util_base.scaling_parameter(temp_par, self.par_scale)
+        rescaled_pars = util.unscale(pars, self.par_scale)
         tree = ET.parse(self.eh.file)  # noqa: S314
         root = tree.getroot()
         for key, val in rescaled_pars.items():
@@ -294,127 +288,101 @@ class MorpheusModel(ExternalModel):
         node = root.findall(xpath)
         return node[0].attrib[attr]
 
-    def compute_sumstats(self, loc):
+    def compute_sumstats(self, loc: str) -> dict:
         """
         Compute summary statistics from the simulated data according to the
         provided list of summary statistics functions.
         """
-        sumstat_dict = {'loc': loc}
-        if "IdSumstat" in self.sumstat_funs[0].name:
-            sumstat_dict = self.sumstat_funs[0](
-                loc, self.ignore_list, self.output_file
-            )
-            sumstat_dict['loc'] = loc
+        sumstat_dict = {LOC: loc}
 
+        if self.sumstat is None:
+            sumstat = SummaryStatistics()
+            # sumstat_dict[LOC] = loc
+            sumstat_dict = sumstat(loc)
             # here add the prepare function,e.g., 0: val,val,val
-            # safe_append_sumstat(sumstat_dict, sumstat, self.sumstat_funs[0])
+            # safe_append_sumstat(sumstat_dict, sumstat,)
             return sumstat_dict
         else:
-            for sumstat_fun in self.sumstat_funs:
-                sumstat = sumstat_fun(loc, self.ignore_list, self.output_file)
-                # here add the prepare function,e.g., 0: val,val,val
-                safe_append_sumstat(sumstat_dict, sumstat, sumstat_fun.name)
-            return sumstat_dict
+            tmp_sumstat = self.sumstat(loc)
+            # here add the prepare function,e.g., 0: val,val,val
+            return tmp_sumstat
 
     def _check_sumstat_funs(self):
         """
         Check sumstat functions for validity.
         """
-        names = [ssf.name for ssf in self.sumstat_funs]
+        names = [ssf.name for ssf in self.sumstat]
         if not len(set(names)) == len(names):
             raise AssertionError(
                 f"The summary statistics passed to MorpheusModel must have"
                 f"unique names, but obtained {names}"
             )
 
-    def _sanity_chceck_model(self):
-        """
-        Sanity check of the model upon construction.
-        """
+    def sanity_check(self, par: Parameter = None):
+        """Sanity check of the model.
 
-        loc = self.eh.create_loc()
-        file_ = os.path.join(loc, "model.xml")
+        In particular executes the model once.
 
-        # write new file with for the model
-        tree = ET.parse(self.eh.file)  # noqa: S314
-        tree.write(file_, encoding="utf-8", xml_declaration=True)
+        Parameters
+        ----------
+        par:
+            Parameters at which to evaluate. If not specified, parameters are
+            as in the model file.
+        """
+        raise_on_error = self.eh.raise_on_error
+        self.eh.raise_on_error = True
+        if par is None:
+            par = Parameter()
+        self(par)
+        self.eh.raise_on_error = raise_on_error
+        logger.info("Sanity check successful")
+
+    def SBML_to_MorpheusML(self,
+                           output_dir=''):
+        SBML_model = os.path.basename(self.eh.file)
+        SBML_model_name, SBML_model_extension = os.path.splitext(SBML_model)
+        if output_dir == "":
+            output_dir = f"./{SBML_model_name}_morpheusML{SBML_model_extension}"
+
+        full_SBML_model = self.eh.file
 
         # create command
-        cmd = self.eh.create_executable(loc)
-        # self.eh.timeout = self.timeout
-        cmd = cmd + f" -file={file_} -outdir={loc}"
+        cmd = self.gui_executable
+        cmd = cmd + f" --convert={full_SBML_model},{output_dir}"
+        self.eh.run(cmd=cmd)
+        self.eh.file = output_dir
+        logger.info("SBML model successfully converted to MorpheusML")
 
-        # call the model
-        status = self.eh.run(cmd=cmd, loc=loc)
-        if status["returncode"] == 0:
-            logger.info("Successfully loaded model")
-        else:
-            raise ValueError(
-                f"The model is not running well. "
-                f"The simulation seems to fail. "
-                f"Returncode: {status['returncode']}"
-            )
-        clean_simulation_output(loc)
+        return self
 
-    def call_post_processing_ss(self, sumstats, function_name='main'):
-        if self.ss_post_processing is not None:
-            if isinstance(self.ss_post_processing, Callable):
-                sumstats = self.ss_post_processing(sumstats)
-            elif isinstance(self.ss_post_processing, dict):
-                if not set(self.ss_post_processing.keys()).issubset(
-                    sumstats.keys()
-                ):
-                    raise ValueError(
-                        "the keys on the 'ss_post_processing' does not "
-                        "match the one on the summary statistics names."
-                    )
-                if isinstance(
-                    list(self.ss_post_processing.values())[0], Callable
-                ):
-                    sumstats = self._call_post_processing_ss_use_function(
-                        sumstats
-                    )
-                elif isinstance(
-                    list(self.ss_post_processing.values())[0], ModuleType
-                ):
-                    sumstats = self._call_post_processing_ss_use_module(
-                        sumstats, function_name
-                    )
-            else:
-                raise ValueError(
-                    f"the type of 'post_processing_ss' should be str or dict."
-                    f" However, {type(self.ss_post_processing)} was given."
-                )
-        return sumstats
+    # def _call_post_processing_ss_use_module(self, sumstats, function_name):
+    #     sumstat_pp = {}
+    #     for key, _module in self.ss_post_processing.items():
+    #         try:
+    #             func = getattr(
+    #                 self.ss_post_processing[key], function_name, None
+    #             )
+    #             sumstat_pp[key] = func({key: sumstats[key]})
+    #         except Exception as e:
+    #             raise RuntimeError(
+    #                 f"the selected ss_post_processing function "
+    #                 f"can not be called. Be sure that the main "
+    #                 f"function called `main(). `{e}"
+    #             )
+    #     return sumstat_pp
 
-    def _call_post_processing_ss_use_module(self, sumstats, function_name):
-        sumstat_pp = {}
-        for key, _module in self.ss_post_processing.items():
-            try:
-                func = getattr(
-                    self.ss_post_processing[key], function_name, None
-                )
-                sumstat_pp[key] = func({key: sumstats[key]})
-            except Exception as e:
-                raise RuntimeError(
-                    f"the selected ss_post_processing function "
-                    f"can not be called. Be sure that the main "
-                    f"function called `main(). `{e}"
-                )
-        return sumstat_pp
-
-    def _call_post_processing_ss_use_function(self, sumstats):
-        sumstat_pp = {}
-        for key, function in self.ss_post_processing.items():
-            try:
-                post_process_ss = function({key: sumstats[key]})
-                sumstat_pp[key] = post_process_ss[key]
-            except Exception as e:
-                raise RuntimeError(
-                    f"the selected ss_post_processing function "
-                    f"can not be called. `{e}"
-                )
-        return sumstat_pp
+    # def _call_post_processing_ss_use_function(self, sumstats):
+    #     sumstat_pp = {}
+    #     for key, function in self.ss_post_processing.items():
+    #         try:
+    #             post_process_ss = function({key: sumstats[key]})
+    #             sumstat_pp[key] = post_process_ss[key]
+    #         except Exception as e:
+    #             raise RuntimeError(
+    #                 f"the selected ss_post_processing function "
+    #                 f"can not be called. `{e}"
+    #             )
+    #     return sumstat_pp
 
 
 class MorpheusModels(ExternalModel):
@@ -425,48 +393,13 @@ class MorpheusModels(ExternalModel):
 
     Parameters
     ----------
-    model_file:
-        The XML file containing the morpheus model.
-    par_map:
-        A dictionary from str to str, the keys being the parameter ids
-        to be used in pyabc, and the values xpaths in the `morpheus_file`.
-    sumstat_funs:
-        List of functions to calculate summary statistics. The list entries
-        are instances of fitmulticell.sumstat.SumstatFun.
-    executable:
-        The path to the morpheus executable. If None given,
-        'morpheus' is used.
-    suffix, prefix:
-        Suffix and prefix to use for the temporary folders created.
-    dir:
-        Directory to put the temporary folders into. The default is
-        the system's temporary files location. Note that these files
-        are usually deleted upon system shutdown.
-    show_stdout, show_stderr:
-        Whether to show or hide the stdout and stderr streams.
-    raise_on_error:
-        Whether to raise on an error in the model execution, or
-        just continue.
-    name:
-        A name that can be used to identify the model, as it is
-        saved to db. If None is passed, the model_file name is used.
-    time_var:
-        The name of the time variable as define in Morpheus model.
-    ignore_list:
-        A list of columns to ignore from Morpheus output. This is introduced to
-        solve the issue with result that cannot be eliminated from morpheus
-        output but yet are not used in the fitting process.
-    timeout:
-        A time in seconds used to early stop Morpheus simulation when exceed.
-    ss_post_processing:
-        A callable function to perform post processing on Morpheus output. If
-        a dict is passed, then specific function will be applied to each
-        summary statistics.
+    models: A list of MorpheusModel objects.
+    name: Name of the joint model.
     """
 
-    def __init__(self, models_list: List[MorpheusModel], name: str = None):
-        self.models_list = models_list
-        self.name = name
+    def __init__(self, models: Sequence[MorpheusModel], name: str = None):
+        self.models: Sequence[MorpheusModel] = models
+        self.name: str = name
 
     def __str__(self):
         s = f"MorpheusModels {{\n" f"\tname      : {self.name}\n" f"}}"
@@ -487,7 +420,7 @@ class MorpheusModels(ExternalModel):
         # a list that will hold result for each experimental condition.
         sumstats_all = {}
         # create target on file system
-        for model in self.models_list:
+        for model in self.models:
             # cond_sumstats = {}
             sumstats = model(pars)
             # for ss_key, ss_val in sumstats.items():
@@ -524,8 +457,7 @@ class MorpheusModels(ExternalModel):
         Write a modified version of the morpheus xml file to the target
         directory.
         """
-        temp_par = copy.deepcopy(pars)
-        rescaled_pars = util_base.scaling_parameter(temp_par, self.par_scale)
+        rescaled_pars = util.unscale(pars, self.par_scale)
         tree = ET.parse(self.eh.file)  # noqa: S314
         root = tree.getroot()
         # parameters
@@ -547,36 +479,36 @@ class MorpheusModels(ExternalModel):
 
         tree.write(file_, encoding="utf-8", xml_declaration=True)
 
-    def call_post_processing_ss(self, sumstats, function_name='main'):
-        if self.ss_post_processing is not None:
-            if isinstance(self.ss_post_processing, Callable):
-                sumstats = self.ss_post_processing(sumstats)
-            elif isinstance(self.ss_post_processing, dict):
-                if not set(self.ss_post_processing.keys()).issubset(
-                    sumstats.keys()
-                ):
-                    raise ValueError(
-                        "the keys on the 'ss_post_processing' does not "
-                        "match the one on the summary statistics names."
-                    )
-                if isinstance(
-                    list(self.ss_post_processing.values())[0], Callable
-                ):
-                    sumstats = self._call_post_processing_ss_use_function(
-                        sumstats
-                    )
-                elif isinstance(
-                    list(self.ss_post_processing.values())[0], ModuleType
-                ):
-                    sumstats = self._call_post_processing_ss_use_module(
-                        sumstats, function_name
-                    )
-            else:
-                raise ValueError(
-                    f"the type of 'post_processing_ss' should be str or dict."
-                    f" However, {type(self.ss_post_processing)} was given."
-                )
-        return sumstats
+    # def call_post_processing_ss(self, sumstats, function_name='main'):
+    #     if self.ss_post_processing is not None:
+    #         if isinstance(self.ss_post_processing, Callable):
+    #             sumstats = self.ss_post_processing(sumstats)
+    #         elif isinstance(self.ss_post_processing, dict):
+    #             if not set(self.ss_post_processing.keys()).issubset(
+    #                 sumstats.keys()
+    #             ):
+    #                 raise ValueError(
+    #                     "the keys on the 'ss_post_processing' does not "
+    #                     "match the one on the summary statistics names."
+    #                 )
+    #             if isinstance(
+    #                 list(self.ss_post_processing.values())[0], Callable
+    #             ):
+    #                 sumstats = self._call_post_processing_ss_use_function(
+    #                     sumstats
+    #                 )
+    #             elif isinstance(
+    #                 list(self.ss_post_processing.values())[0], ModuleType
+    #             ):
+    #                 sumstats = self._call_post_processing_ss_use_module(
+    #                     sumstats, function_name
+    #                 )
+    #         else:
+    #             raise ValueError(
+    #                 f"the type of 'post_processing_ss' should be str or dict."
+    #                 f" However, {type(self.ss_post_processing)} was given."
+    #             )
+    #     return sumstats
 
     def _call_post_processing_ss_use_module(self, sumstats, function_name):
         sumstat_pp = {}
@@ -608,27 +540,27 @@ class MorpheusModels(ExternalModel):
         return sumstat_pp
 
 
-def safe_append_sumstat(sumstat_dict, sumstat, key):
-    types_ = (numbers.Number, np.ndarray, pd.DataFrame)
-    if isinstance(sumstat, types_):
-        if key in sumstat_dict:
-            raise KeyError(
-                f"Key {key} for sumstat {sumstat} already in the "
-                f"sumstat dict {sumstat_dict}."
-            )
-        sumstat_dict[key] = sumstat
-        return
-    if isinstance(sumstat, dict):
-        if key in sumstat_dict:
-            raise KeyError(
-                f"Key {key} for sumstat {sumstat} already in the "
-                f"sumstat dict {sumstat_dict}."
-            )
-        sumstat_dict.update(sumstat)
-        return
-    raise ValueError(
-        f"Type {type(sumstat)} of sumstat {sumstat} " f"is not permitted."
-    )
+# def safe_append_sumstat(sumstat_dict, sumstat):
+#     types_ = (numbers.Number, np.ndarray, pd.DataFrame)
+#     if isinstance(sumstat, types_):
+#         if key in sumstat_dict:
+#             raise KeyError(
+#                 f"Key {key} for sumstat {sumstat} already in the "
+#                 f"sumstat dict {sumstat_dict}."
+#             )
+#         sumstat_dict[key] = sumstat
+#         return
+#     if isinstance(sumstat, dict):
+#         if key in sumstat_dict:
+#             raise KeyError(
+#                 f"Key {key} for sumstat {sumstat} already in the "
+#                 f"sumstat dict {sumstat_dict}."
+#             )
+#         sumstat_dict.update(sumstat)
+#         return
+#     raise ValueError(
+#         f"Type {type(sumstat)} of sumstat {sumstat} " f"is not permitted."
+#     )
 
 
 def clean_simulation_output(loc):
