@@ -3,21 +3,13 @@ Summary statistics
 ==================
 
 A library of summary statistics common in applications with
-multi-celluar systems.
+multi-cellular systems.
 
 """
 
-from .base import IdSumstatFun, SumstatFun
+from .base import SummaryStatistics
 from .cell_types_cout import CellCountSumstatFun
 from .hexagonal_cluster_sumstat import (
     CCContributorsAllTpCountSumstatFun,
     ClusterCountSumstatFun,
 )
-
-__all__ = [
-    'SumstatFun',
-    'IdSumstatFun',
-    'ClusterCountSumstatFun',
-    'CCContributorsAllTpCountSumstatFun',
-    'CellCountSumstatFun',
-]
